@@ -1,8 +1,10 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-from trytond.model import fields
+from sql import Null
+
+from trytond.model import Check, fields
+from trytond.pool import Pool, PoolMeta
 from trytond.pyson import Eval
-from trytond.pool import PoolMeta
 
 
 class Location(metaclass=PoolMeta):
@@ -14,8 +16,7 @@ class Location(metaclass=PoolMeta):
             },
         domain=[
             ('type', '=', 'production'),
-            ],
-        depends=['type'])
+            ])
     production_picking_location = fields.Many2One(
         'stock.location', "Production Picking",
         states={
@@ -23,9 +24,8 @@ class Location(metaclass=PoolMeta):
             },
         domain=[
             ('type', '=', 'storage'),
-            ('parent', 'child_of', [Eval('id')]),
+            ('parent', 'child_of', [Eval('id', -1)]),
             ],
-        depends=['type', 'id'],
         help="Where the production components are picked from.\n"
         "Leave empty to use the warehouse storage location.")
     production_output_location = fields.Many2One(
@@ -35,29 +35,57 @@ class Location(metaclass=PoolMeta):
             },
         domain=[
             ('type', '=', 'storage'),
-            ('parent', 'child_of', [Eval('id')]),
+            ('parent', 'child_of', [Eval('id', -1)]),
             ],
-        depends=['type', 'id'],
         help="Where the produced goods are stored.\n"
         "Leave empty to use the warehouse storage location.")
 
 
 class Move(metaclass=PoolMeta):
     __name__ = 'stock.move'
-    production_input = fields.Many2One('production', 'Production Input',
-        readonly=True, select=True, ondelete='CASCADE',
+    production_input = fields.Many2One(
+        'production', "Production Input", readonly=True, ondelete='CASCADE',
         domain=[('company', '=', Eval('company'))],
-        depends=['company'])
-    production_output = fields.Many2One('production', 'Production Output',
-        readonly=True, select=True, ondelete='CASCADE',
+        states={
+            'invisible': ~Eval('production_input'),
+            })
+    production_output = fields.Many2One(
+        'production', "Production Output", readonly=True, ondelete='CASCADE',
         domain=[('company', '=', Eval('company'))],
-        depends=['company'])
+        states={
+            'invisible': ~Eval('production_output'),
+            })
+    production = fields.Function(fields.Many2One(
+            'production', "Production",
+            states={
+                'invisible': ~Eval('production'),
+                }),
+        'on_change_with_production')
     production_cost_price_updated = fields.Boolean(
         "Cost Price Updated", readonly=True,
         states={
             'invisible': ~Eval('production_input') & (Eval('state') == 'done'),
-            },
-        depends=['production_input', 'state'])
+            })
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        t = cls.__table__()
+        cls._sql_constraints += [
+            ('production_single', Check(t, (
+                        (t.production_input == Null)
+                        | (t.production_output == Null))),
+                'production.msg_stock_move_production_single'),
+            ]
+
+    @fields.depends(
+        'production_input', '_parent_production_input.id',
+        'production_output', '_parent_production_output.id')
+    def on_change_with_production(self, name=None):
+        if self.production_input:
+            return self.production_input
+        elif self.production_output:
+            return self.production_output
 
     def set_effective_date(self):
         if not self.effective_date and self.production_input:
@@ -96,3 +124,44 @@ class ProductQuantitiesByWarehouseMove(metaclass=PoolMeta):
         if self.move.production_output:
             document = str(self.move.production_output)
         return document
+
+
+class LotTrace(metaclass=PoolMeta):
+    __name__ = 'stock.lot.trace'
+
+    production_input = fields.Many2One('production', "Production Input")
+    production_output = fields.Many2One('production', "Production Output")
+
+    @classmethod
+    def _columns(cls, move):
+        return super()._columns(move) + [
+            move.production_input.as_('production_input'),
+            move.production_output.as_('production_output'),
+            ]
+
+    @classmethod
+    def get_documents(cls):
+        pool = Pool()
+        Model = pool.get('ir.model')
+        return super().get_documents() + [
+            ('production', Model.get_name('production'))]
+
+    def get_document(self, name):
+        document = super().get_document(name)
+        if self.production_input:
+            document = str(self.production_input)
+        elif self.production_output:
+            document = str(self.production_output)
+        return document
+
+    def _get_upward_traces(self):
+        traces = super()._get_upward_traces()
+        if self.production_input:
+            traces.update(self.production_input.outputs)
+        return traces
+
+    def _get_downward_traces(self):
+        traces = super()._get_downward_traces()
+        if self.production_output:
+            traces.update(self.production_output.inputs)
+        return traces
